import polars as pl
from collections import namedtuple
from rich import print

"""
##############################################################################
### Loading FPKM data for Homo sapiens. 
def reloadFPKMHsapiens(fpkmDic, model):
    newfpkmDic = {}
    for gene in model.genes:
        if not 'G_'+gene.name in fpkmDic: continue
        fpkm = fpkmDic['G_'+gene.name]
        gene = 'G_'+gene.id
        newfpkmDic[gene] = fpkm
    return(newfpkmDic)

"""

HumanGene = namedtuple("HumanGene", ["hid", "Gene_ID"])


def reload_fpkmh_sapiens(
        fpkm:pl.DataFrame, 
        model:pl.DataFrame):
    """
    Match genes that exists on 'fpkm'

    caso excepcional
    """
    genes = [HumanGene(f"G_{gene.id}", f"G_{gene.name}") for gene in model.genes]
    hg_dataframe = pl.DataFrame(genes)
    joined_df = fpkm.join(hg_dataframe, on="Gene_ID", how="inner")
    df = joined_df.drop('Gene_ID')
    df = df.rename({"hid":"Gene_ID"})
    return df



from cobra.core.model import Model as SBMLModel

def getForwardReverse(model):
    """
    Input:
        model (cobra.Model): Metabolic model.
    
    Output:
        tuple: Lists of forward and reverse variable names.
    """
    v_vars, rev_vars = [], []
    for reaction in model.reactions:
        v_vars.append(reaction.id)
        rev_vars.append(reaction.reverse_id)  
    return(v_vars,rev_vars)



from casadi import vertcat, SX, log
from collections import namedtuple
from cobra.core.model import Model as SBMLModel
import polars as pl
from pheflux.get_eg import get_gene_expression, BIAS
from gene_rule_parser import (
    ExpressionParser, 
    tree_analysis, 
    Gene, 
    distribute_rules)
from dataclasses import field
from pydantic.dataclasses import dataclass

CasADIData = namedtuple("CasADIData", ["casadi", "names","variables",
                                       "lower_bounds", "upper_bounds", 
                                       "objetive_fn", "sx_variables", 
                                       "primal"]) 


@dataclass(config=dict(arbitrary_types_allowed=True))
class ReactionVariables:
    variables:dict[str, SX] = field(default_factory=dict)

    def set(self, key, value):
        self.variables[key] = value
    
    def get(self, key):
        return self.variables.get(key)


def set_variables(
        model:SBMLModel, 
        fpkm:pl.DataFrame
)->CasADIData:
    """
    Input:
        model (cobra.Model): Metabolic model.
        fpkmDic (dict): Dictionary with gene expression values.
    
    Output:
        tuple: CasADI object with all variables, dictionary with CasADI variables, lists of lower and upper bounds, and objective function.
    """
    
    v = vertcat() # saves the total of variables of the model. Used to """nlp['x']"""
    v_dic = {}
    v_fpkm = {} # 
    fba_primal = {}
    ubx, lbx = [],[]
    reaction_names = []
    ##############################################################
    ## Gets the median value of 'g'
    parser = ExpressionParser()
    # Esperanza(g), g_metab (booleans by rule [dict])

    
    gen_metab = get_gene_expression(parser, model, fpkm)     
    r_vars = ReactionVariables()
    counter = 0

    for i, reaction in enumerate(model.reactions):
        rule = reaction.gene_reaction_rule
        # dar expr genética = promedio de las conocidas
        fba_primal[reaction.id] = getattr(model.variables,
                                         reaction.id).primal
        fba_primal[reaction.reverse_id] = getattr(model.variables,
                                                  reaction.reverse_id).primal

        var_name = reaction.id
        reaction_id = var_name
        var_name = SX.sym(var_name)
        v = vertcat(v, var_name)

        var_name_reverse = reaction.reverse_id
        var_name_reverse = SX.sym(var_name_reverse)
        v = vertcat(v,var_name_reverse)
        
        # add primals, get default
        g = gen_metab.get(rule)
        
        #reaction = gen_metab.get_id(rule)
        
        r_vars.set(reaction.id, var_name)
        v_dic[reaction_id] = var_name

        ubx.append(reaction.upper_bound)    
        lbx.append(0.0)

        # reverse
        r_vars.set(reaction.reverse_id, var_name_reverse)
        v_dic[reaction.reverse_id] = var_name_reverse

        ubx.append(-reaction.lower_bound)
        lbx.append(0.0)

        #
        reaction_names.append((reaction.id, reaction.reverse_id))
        # calc
        v_fpkm[var_name] = g
        v_fpkm[var_name_reverse] = g

        ##########################################################
        ## Define a objective function
        for j, name in enumerate([var_name,var_name_reverse]):
            if counter == 0 and j==0:
                v_ViLogVi = ( (name)+BIAS )*log( (name)+BIAS ) # 1.1
                v_VilogQi = ( (name)+BIAS )*log( g ) # 2.1
            else:
                v_ViLogVi += ( (name)+BIAS )*log( (name)+BIAS ) # 1.1
                v_VilogQi += ( (name)+BIAS )*log( g ) # 2.1            
        counter += 1
    ##############################################################
    ## Set objetive function
    f = (v_ViLogVi) - (v_VilogQi)
    # v is ---> x variables
    # f is ---> f objetive function
    return CasADIData(v,reaction_names, v_dic,lbx,ubx,f, r_vars, fba_primal)
    #gen_metab: contains list of genes and values, and have property method: mean
#################################################################################
### Variables and objective function: CasADI object
"""
def setVariables(model,fpkmDic):
    v = vertcat() 
    v_dic = {}
    v_fpkm = {} # 
    ubx, lbx = [],[]
    ##############################################################
    ## Gets the median value of 'g'
    E_g,g_metab = getEg(model,fpkmDic)     
    
    for i, reaction in enumerate(model.reactions):
        ##########################################################
        ## Gets the GPR and a boolean list with known or unknown genes
        rule = reaction.gene_reaction_rule # gene reaction rule
        boolean_list = booleanVectorRule(rule,fpkmDic) # useful to discriminate between genes with known fkpm.
        ##########################################################
        ## Gets the expression value 'g'
        # get 'g' for reaction with GPR.
        if not ('False' in boolean_list or rule == ''): 
            g = getG(rule, fpkmDic)+1e-8#*1e-8
        # set 'g' (median value) for reaction without GPR. 
        else:
            g = E_g
        ##########################################################
        ## Set forward and reverse variables as a CasADI object
        # forward
        var_name = reaction.id
        expression = var_name+' = SX.sym("'+var_name+'")'
        exec(expression, globals())
        vf = eval(var_name)
        v = vertcat(v, vf)
        v_dic[reaction.id]=vf
        ubx.append(reaction.upper_bound)    
        lbx.append(0.0)

        # reverse
        var_name_reverse = reaction.reverse_id
        expression = var_name_reverse+' = SX.sym("'+var_name_reverse+'")'
        exec(expression, globals())
        vr = eval(var_name_reverse)
        v = vertcat(v,vr)
        v_dic[reaction.reverse_id]=vr
        ubx.append(-reaction.lower_bound)
        lbx.append(0.0)

        v_fpkm[var_name] = g
        v_fpkm[var_name_reverse] = g
        ##########################################################
        ## Define a objective function
        for name in [vf,vr]:
            if i == 0:
                v_ViLogVi = ( (name)+1e-8 )*log( (name)+1e-8 ) # 1.1
                v_VilogQi = ( (name)+1e-8 )*log( g ) # 2.1
            else:
                v_ViLogVi += ( (name)+1e-8 )*log( (name)+1e-8 ) # 1.1
                v_VilogQi += ( (name)+1e-8 )*log( g ) # 2.1            
    ##############################################################
    ## Set objetive function
    f = (v_ViLogVi) - (v_VilogQi)
    return(v,v_dic,lbx,ubx,f)
"""



import tomllib
import yaml
from typing import Union, Any, Dict, List
from pathlib import Path
# local
from pheflux.item import OrganismItem
from pydantic import BaseModel, validator

class Settings(BaseModel):
    name: str
    author: str
    verbosity: bool    
    organisms: List[OrganismItem]


PATH_FIELDS = {"gene_exp_file","medium","network"}

def set_paths(base_path, item:Dict[str,str]):
    for field in PATH_FIELDS:
        value = item[field]
        item[field] = f"{base_path}/{value}"
    return item 

def read_file(file_path: Path) -> Settings|None:
    """
    Lee configuraciones desde un archivo TOML o YAML.
    
    :param file_path: Ruta del archivo de configuración como una instancia de Path.
    :return: Diccionario con las configuraciones, o None si no se pudo leer el archivo.
    """
    try:
        base_path = file_path.parent.parent.resolve()        
        if file_path.suffix == '.toml':
            with file_path.open('rb') as file:
                settings_dict = tomllib.load(file)
                result = [set_paths(base_path, r) for r in settings_dict.pop("input_data")]
                return Settings(organisms=[OrganismItem(**item) for
                                           item in result], **settings_dict)
        elif file_path.suffix in {'.yaml', '.yml'}:
            with file_path.open('r') as file:
                settings_dict = yaml.safe_load(file)
                result = [set_paths(base_path, r) for r in settings_dict.pop("input_data")]
                return Settings(organisms=[OrganismItem(**item) for
                                           item in result], **settings_dict)
        else:
            print("Formato de archivo no soportado. Use un archivo .toml o .yaml/.yml.")
            return None
    except (FileNotFoundError, yaml.YAMLError, tomllib.TOMLDecodeError) as e:
        print(f"Error al leer el archivo de configuración: {e}")
        return None




import polars as pl
import ply.yacc as yacc
from gene_rule_parser import (
    ExpressionParser, 
    tree_analysis, 
    Gene, 
    distribute_rules)

##############################################################################
### Obtains expression value 'g' for each reaction, based on the gene-protein-rule
def get_expression_value(
        tree:yacc.LRParser,
        fpkm:pl.DataFrame)->float:

    """
    Input:
        tree (str): Gene-protein tree parsed.
        fpkmDic (dict): Dictionary with gene expression values.
    
    Output:
        float: Sum of minimum expression values for each subrule in the gene-protein rule.
    """
    return tree_analysis(tree, fpkm)



from cobra.core.model import Model as SBMLModel
from casadi import vertcat, SX, log

def is_number(s: str) -> bool:
    try:
        float(s)  # for int, float, and complex numbers
        return True
    except ValueError:
        return False

def createConstraints(
        model:SBMLModel,
        reaction_names:list[tuple[str,str]],
        k:float,
        v_dic,
        sumVi:SX, 
        sx_variables):
    """
    Input:
        model (cobra.Model): Metabolic model.
        k (float): SumV constraint value.
        v_dic (dict): Dictionary with CasADI variables.
        sumVi (CasADI object): Sum of all variables.
    
    Output:
        tuple: Lists of upper and lower bounds for constraints, and CasADI object with all constraints.
    """
    g = vertcat()
    lbg,ubg=[],[]
    ##############################################################
    ## Gets the name of the forward/reverse variables
    v_vars, rev_vars = zip(*reaction_names)
    v_vars = list(v_vars)
    rev_vars = list(rev_vars)
    ##############################################################
    ## Defines constraints
    discard = False
    ops = {"*","-","+"}
    for met in model.metabolites:
        ##########################################################
        ## Gets constraint for a one metabolite
        elements = str(met.constraint.expression).replace("*"," * ").split()
        sx_expression = []
        mults = []
        for im, e in enumerate(elements):
            if e=="*":
                mults.append(im)
            if is_number(e):
                e = float(e)
            
            elif e not in ops and not is_number(e):
                e_ = sx_variables.get(e)
                if e_ is None:
                    discard = True
                else:
                    e = e_
            sx_expression.append(e)
        #breakpoint()
        #constraint = str(met.constraint.expression).replace('+ ','+').replace('- ','-')
        ##########################################################
        ## Reconstruct the constraint as a CasADI object
        post_sx_expression = []
        new_muls = []
        if not discard:
            for im in mults:
                pre = im-1
                post = im+1
                pre_val = sx_expression[pre]
                post_val = sx_expression[post]
                post_sx_expression.append(pre_val*post_val)
                next_symbol = im+2
                if next_symbol<len(sx_expression):
                    symbol = sx_expression[next_symbol]
                    post_sx_expression.append(f"{symbol}")

            for m,value in enumerate(post_sx_expression):
                if m==0:
                    tmp_constraint = value 

                if value in {"-","+"}:
                    pre_value = post_sx_expression[m-1]
                    post_value = post_sx_expression[m+1]
                    
                    match value:
                        case "-":
                            tmp_constraint -= post_value
                        case "+":
                            tmp_constraint += post_value

            ##########################################################
            ## Adds constraint to CasADI
            g = vertcat(g,tmp_constraint)
            lbg.append(0)
            ubg.append(0)
        discard = False
    ##############################################################
    ## SumV constraint
    g = vertcat(g,sumVi)
    lbg.append( k )
    ubg.append( k )

    return ubg,lbg,g



from cobra.core.model import Model as SBMLModel
import json
from pathlib import Path

BASE  = Path(__file__).parent.parent.resolve() 

def save_rules(model:SBMLModel, name:str):
    rules = set()
    for i, reaction in enumerate(model.reactions):
        rule = reaction.gene_reaction_rule.strip()
        rules.add(rule)

    txt = json.dumps(list(rules))
    p = BASE / f"{name}_rules.json"
    p.write_text(txt)



# from std
from pathlib import Path
from dataclasses import dataclass, field
from datetime import datetime,timezone
from typing import Optional
import random
import string
import time
import zipfile

# contrib
import polars as pl
import pandas as pd
import numpy as np
import typer
from rich import print
import ujson as json

# local
from pheflux.item import OrganismItem
from pheflux.measure_time import MeasureTime
from pheflux.read_settings import read_file, Settings
from pheflux.utils import actual_time
from pheflux.load_fpkm import load_fpkm
from pheflux.reload_fpkmh_sapiens import reload_fpkmh_sapiens
from pheflux.update_model import update_model
from pheflux.opt_phe_flux import opt_phe_flux
from pheflux.save_rules import save_rules
from pheflux.record import record_table

app = typer.Typer()

#################################################################################
########################             PHEFLUX             ########################
#################################################################################
"""
"""

@dataclass
class ResultItem:
    organism:str
    condition:str
    csv_file:Path|None = None

    def set_csv(self, csv_file:Path):
        if csv_file.exists():
            self.csv_file = csv_file

@dataclass
class ResultGroup:
    name:str
    log:Path 
    results:list[ResultItem] = field(default_factory=list)
    report:dict[str, float] = field(default_factory=dict)
    
    def add(self, item:ResultItem):
        self.results.append(item)


    def set_report(self, report:dict[str, float]):
        self.report = report
        

    def generate_zip(self, output_zip: Path):

        now = datetime.now(timezone.utc).isoformat()
        
        if not output_zip.is_dir():
            output_zip = output_zip.parent
        output_zip.mkdir(exist_ok=True, parents=True)
            

        output_zip = output_zip / f"{self.name}_{now}.zip"
        print(f"Zip output: {output_zip}")
        
        with zipfile.ZipFile(output_zip, 'w') as zipf:
            # Add log file to the zip
            if self.log.exists():
                zipf.write(self.log, arcname=self.log.name)
            
            # Add each csv_file from ResultItem to the zip
            for result_item in self.results:
                if result_item.csv_file and result_item.csv_file.exists():
                    zipf.write(result_item.csv_file, arcname=result_item.csv_file.name)
            
            # Create a JSON file for the report and add it to the zip
            report_json_path = Path('report_time.json')
            with report_json_path.open('w') as json_file:
                json.dump(self.report, json_file)
            zipf.write(report_json_path, arcname=report_json_path.name)
            
            # Clean up the temporary report JSON
            report_json_path.unlink(missing_ok=True)
        return output_zip

    
def get_fluxes(
        settings:Settings, task_name:str=""
):
    """
    Input:
        inputFileName (str): File name with input data.
        processDir (str): Directory for saving process files.
        prefix_log (str): Prefix for log files.
        verbosity (bool): Flag to control verbosity of the process.
    
    Output:
        Series: Series with fluxes.
    """
    processDir = "process"
    prefix_log = 'log'
    
    fluxes = None
    process_start = time.time()
    shuffle=False
    shuffled_fpkm = pl.DataFrame()
    record = pl.DataFrame()

    opt_time, t_time = [], []
    code = ''.join(
        random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(4)).upper()
    
    recordFile = Path(f"{processDir}/{prefix_log}_record_{code}.log.csv")
    
    results_group = ResultGroup(task_name, recordFile)
    # operate this
    for item in settings.organisms:
        item.activate()
        condition    = item.condition
        prefix_log   = str(condition) 

        geneExpFile  = item.gene_exp_file
        mediumFile   = item.medium
        medium_data  = item.medium_data
        network      = item.network
        organism     = item.organism

        result_item = ResultItem(organism, condition)
        ##############################################################
        ## Messages in terminal
        atime = actual_time()
        print(atime, 'Condition ejecuted:', organism, '-', condition)
        ##############################################################
        # Metabolic network
        if settings.verbosity:
            atime = actual_time()
            print (atime,"Loading metabolic model:", network.stem)
        model_default = item.network_data
        fpkm = item.gene_exp_data

        print("Model")
        print(model_default)
        print("____")
        init_time = time.time()
        ##############################################################
        # FPKM data
        if settings.verbosity:
            atime = actual_time()
            print(atime, "Loading transcriptomic data...")
        # Load FPKM data
        # fpkmDic,shuffled_fpkm = load_fpkm(
        #     fpkm,
        #     condition,
        #     shuffle,
        #     shuffled_fpkm)
        fpkm_filtered = load_fpkm(fpkm)


        # # Reload FPKM data for Hsapiens and load culture medium
        # if organism == 'Homo_sapiens':
        #     fpkmDic = reloadFPKMHsapiens(fpkmDic, model_default)
        if organism.lower() == 'homo_sapiens':
            fpkm_filtered = reload_fpkmh_sapiens(fpkm_filtered, model_default)
            print(fpkm_filtered)

        ##############################################################
        # Update model: Add R_, open bounds, and set carbon source
        if settings.verbosity:
            atime = actual_time()
            print(atime, "Updating metabolic model...")
        model = update_model(model_default, medium_data)


        ##############################################################
        # Compute flux predictions
        if settings.verbosity:
            atime = actual_time()
            print(atime, "Running pheflux...")
        k = 1000

        # in case to save rules to create some tests:
        #save_rules(model, organism)
        # checked here()

        fluxes,optimization_time,total_time,status,success,lbx,ubx = opt_phe_flux(
            model,
            fpkm_filtered,
            init_time,
            k)

        ##############################################################
        # Save results: fluxes and summary table
        print(" ")
        if settings.verbosity:
            atime = actual_time()
            print(atime, "Saving metabolic fluxes...")
        # fluxes
        resultsFile = Path(f"{processDir}/{organism}_{condition}_{status}.fluxes.csv")
        result_item.set_csv(resultsFile)
        fluxes.write_csv(resultsFile, separator=';')
        # summary table
        record = record_table(record,condition,lbx,ubx,total_time,status)
        
        ##############################################################
        ## Messages in terminal
        opt_time.append(optimization_time)
        t_time.append(total_time)
        atime = actual_time()
        print(atime, organism, '-', condition, "... is processed.")

        print ('\n',' o '.center(80, '='),'\n')
        results_group.add(result_item)


    ##############################################################
    # Save summary table
    
    # create directory if necessary
    processPath = Path(processDir)
    processPath.mkdir(exist_ok=True, parents=True)
    record.write_csv(recordFile, separator=';')

    process_end = time.time()

    measure_time = MeasureTime(
        start_time=process_start,
        end_time=process_end,
        opt_time=opt_time,
        t_time=t_time
    )
    report_json = measure_time.generate_report()
    results_group.set_report(report_json)
    return fluxes, report_json, results_group



@app.command()
def run(settings_path:Path):
    if settings_path.exists():
        settings = read_file(settings_path)
        print(settings)
        fluxes, report, results = get_fluxes(settings, "test_pheflux")
        print(report)
        print(results)
        out_dir = Path("outputs")
        out_dir.mkdir(exist_ok=True, parents=True)
        results.generate_zip(out_dir)
    else:
        print(f"The settings path doesn't exists {settings_path}")

if __name__=="__main__":
    app()



##############################################################################
### Loading FPKM data
"""
def loadFPKM(fpkm,condition,shuffle=False,shuffledFPKM=pd.DataFrame()):
    ##########################################################
    ## Gets gene IDs and their expression values
    genes=fpkm["Gene_ID"]
    if shuffle:
        fpkms=fpkm["Expression"].sample(frac=1).reset_index(drop=True) 
    else:
        fpkms=fpkm["Expression"]
    shuffledFPKM["Expression"] = fpkms
    ##########################################################
    ## Creates a dictionary gene_ID -> Expression value
    fpkmDic = {}
    for i in range(len(fpkms)): # Run over each line in fpkm file
        # 1. Get gene id and fpkm values for each line
        name = 'G_'+str(genes[i])
        fpkm = fpkms[i]
        if type(fpkm) == np.float64:
            fpkmDic[name] = fpkm
    ##########################################################
    ## Capping at 95%
    cap = np.percentile( list(fpkmDic.values()), 95)
    for i in fpkmDic:
        if fpkmDic[i]>cap:
            fpkmDic[i] = cap
    return(fpkmDic,shuffledFPKM)
"""

import polars as pl

def load_fpkm(
    fpkm:pl.DataFrame,
):

    
    fpkm = fpkm.with_columns(
        pl.col("Gene_ID").map_elements(
            lambda x: f"G_{x}", return_dtype=pl.Utf8).alias("Gene_ID")
    )
    
    fpkm = fpkm.unique(subset=["Gene_ID"], keep='last')
    
    expression = "Expression"

    
    percentile_95 = fpkm[expression].quantile(0.95)
    #filtered_df = fpkm.filter(pl.col(expression) <= percentile_95)

    filtered_df = fpkm.with_columns(
        pl.col(expression).map_elements(lambda x: min(x, percentile_95),
                                        return_dtype=pl.Float64
                                        ).alias(expression)
    )    

    return filtered_df



import pandas as pd
import polars as pl

#################################################################################
### Table of times and variable numbers for networks
def record_table(record ,condition, lbx, ubx, time, status):   
    """
    Input:
        record (DataFrame): DataFrame with records of previous runs.
        condition (str): Condition name.
        lbx (list): List of lower bounds.
        ubx (list): List of upper bounds.
        time (float): Total time of the run.
        status (str): Status of the run.
    
    Output:
        DataFrame: Updated record DataFrame.
    """
    variables = 0
    for i in range(len(lbx)):
        if lbx[i] != ubx[i]:
            variables += 1
    
    if record.shape == (0,0):
        record = pl.DataFrame(schema={'Condition':str, 'N° variables':int, 'Time':float, 'Status':str})

    item = {
        "Condition":condition,
        "N° variables":variables,
        "Time":time,
        "Status":status
    }
    new_record = pl.DataFrame(item)
    record = pl.concat([record, new_record])
    return record




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'leftORleftANDAND GENE LPAREN OR RPAREN UNKNOWNexpression : expression OR expressionexpression : expression AND expressionexpression : LPAREN expression RPARENexpression : GENEexpression : UNKNOWN'
    
_lr_action_items = {'LPAREN':([0,2,5,6,],[2,2,2,2,]),'GENE':([0,2,5,6,],[3,3,3,3,]),'UNKNOWN':([0,2,5,6,],[4,4,4,4,]),'$end':([1,3,4,8,9,10,],[0,-4,-5,-1,-2,-3,]),'OR':([1,3,4,7,8,9,10,],[5,-4,-5,5,-1,-2,-3,]),'AND':([1,3,4,7,8,9,10,],[6,-4,-5,6,6,-2,-3,]),'RPAREN':([3,4,7,8,9,10,],[-4,-5,10,-1,-2,-3,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expression':([0,2,5,6,],[1,7,8,9,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expression","S'",1,None,None,None),
  ('expression -> expression OR expression','expression',3,'p_expression_or','gen_parser.py',40),
  ('expression -> expression AND expression','expression',3,'p_expression_and','gen_parser.py',44),
  ('expression -> LPAREN expression RPAREN','expression',3,'p_expression_group','gen_parser.py',48),
  ('expression -> GENE','expression',1,'p_expression_gene','gen_parser.py',52),
  ('expression -> UNKNOWN','expression',1,'p_expression_unknown','gen_parser.py',56),
]



import polars as pl
from collections import namedtuple
from rich import print
from cobra.core.model import Model as SBMLModel

"""
##############################################################################
# UPDATE MODEL
def updateModel(model_default,mediumFile):
    model=model_default.copy()
    ##########################################################
    ## Add 'R_' to reactions names
    for reaction in model.reactions:
        reaction.id = 'R_'+reaction.id
    ##########################################################        
    ## Opening the model: exchange reactions
    for rxn in model.reactions: 
        if (rxn.lower_bound<0 and rxn.upper_bound>0):
            rxn.bounds = (-1000,1000)
        if (rxn.lower_bound>=0 and rxn.upper_bound>0):
            rxn.bounds = (0,1000)
        if (rxn.lower_bound<0 and rxn.upper_bound<=0):
            rxn.bounds = (-1000,0)
    ##########################################################
    ## Set culture medium
    #############################################
    if mediumFile != 'NA':
        #####################
        # load medium
        medium = pd.read_csv(mediumFile,sep="\t", lineterminator='\n')
        #####################
        # set bounds of exchange reactions to (0, 1000)
        for reaction in model.exchanges:
            reaction.bounds = (0, 1000)
        #####################
        # add culture medium
        for reaction in medium['Reaction_ID']:
            if 'R_'+reaction in model.reactions:
                model.reactions.get_by_id('R_'+reaction).lower_bound = -1000
        
    return(model)


"""

def update_model(
        model_default:SBMLModel,
        medium:pl.DataFrame):
    """
    Input:
        model_default (cobra.Model): Original metabolic model.
        mediumFile (str): File name of the medium conditions.
    
    Output:
        cobra.Model: Updated metabolic model.
    """
    model = model_default.copy()
    ##########################################################
    ## Add 'R_' to reactions names
    for rxn in model.reactions:
        rxn.id = f"R_{rxn.id}"
        lower = rxn.lower_bound
        upper = rxn.upper_bound
        match lower, upper:
            case (v1,v2) if v1<0 and v2>0:
                rxn.bounds = (-1000,1000)
            case (v1,v2) if v1>=0 and v2>0:
                rxn.bounds = (0,1000)
            case (v1,v2) if v1<0 and v2<=0:
                rxn.bounds = (-1000,0)
    ##########################################################
    ## Set culture medium
    #############################################
    if not medium.is_empty():
        for reaction in model.exchanges:
            reaction.bounds = (0, 1000)

        unique_reactions = medium.select(pl.col("Reaction_ID").unique()).to_series().to_list()

        for reaction in unique_reactions:
            name = f"R_{reaction}"
            if name in model.reactions:
                model.reactions.get_by_id(name).lower_bound = -1000


    return model



import polars as pl
import numpy as np
from cobra.core.reaction import Reaction
from cobra.core.model import Model as SBMLModel
from collections import namedtuple
from pheflux.boolean_vector_rule import boolean_vector_rule
from pheflux.get_g import get_expression_value
from gene_rule_parser import (ExpressionParser, tree_analysis)
from collections import namedtuple
from pydantic.dataclasses import dataclass
from functools import lru_cache
##############################################################################
## Obtains a median value of the expression of genes associated with
## metabolism
import ujson
from pathlib import Path

@dataclass(config=dict(arbitrary_types_allowed=True))
class GenMetab:
    reaction_ids:dict[str,Reaction]
    rules:dict[str, float]
    booleans:dict[str,dict[str, bool]]

    @property
    def mean(self)->float:
        return float(np.mean(list(self.rules.values())))


    def get(self, rule:str)->float:
        return self.rules.get(rule, self.mean)

    def get_id(self, rule:str)->str:
        return self.reaction_ids.get(rule)


BIAS = 1e-8
def get_gene_expression(
        parser:ExpressionParser,
        model:SBMLModel,
        fpkm:pl.DataFrame
)->GenMetab:
    """
    Input:
        model (cobra.Model): Metabolic model.
        fpkmDic (dict): Dictionary with gene expression values.
    
    Output:
        tuple: Median value of gene expressions, and list of gene expressions for metabolic reactions.
    """
    g_metab = {} # gene expression of reactions partakin in the
    # metabolism
    #parser = ExpressionParser()
    rules_cache = {}
    genes_values = {}
    booleans = {}
    reaction_ids = {}

   
    
    for i, reaction in enumerate(model.reactions):
        ##########################################################
        ## Gets the GPR and a boolean list with known or unknown genes
        rule = reaction.gene_reaction_rule.strip()
        if rule!='':
            reaction_ids[rule] = reaction
            check_bool = rules_cache.get(rule, False)

            if not check_bool:
                tokens = parser.tokenize(rule)
                rule_booleans = boolean_vector_rule(tokens, fpkm)
                booleans[rule] = rule_booleans
                rules_cache[rule] = all(rule_booleans.values()) and len(rule_booleans)>0
            ##########################################################
            ## Gets the expression value 'g'
            ## for every exists rule, add g
            check_bool = rules_cache.get(rule, False)
            if check_bool: # get 'g' for reaction with GPR.
                if rule not in genes_values:
                    rule_tree = parser.parse(rule)
                    g = get_expression_value(rule_tree, fpkm)
                    genes_values[rule] = g
                else:
                    g = genes_values[rule]
                g_metab[rule]=g+BIAS
    return GenMetab(reaction_ids, g_metab, booleans)



#!/usr/bin/env python3

from casadi import *
import cobra

import pandas as pd
import numpy as np
import time

from datetime import datetime
import random
import string

from pheflux import getG

##############################################################################
### This function gives a useful vector to discriminate if the FPKM of the
### genes associated with a reaction is known.
def booleanVectorRule (rule, fpkmDic):
    boolean_list = []
    vector_rule = rule.replace("or","")
    vector_rule = vector_rule.replace("and","")
    vector_rule = vector_rule.replace("'","") # Removes "'"
    vector_rule = vector_rule.replace("(","") # Remove "("
    vector_rule = vector_rule.replace(")","")
    vector = vector_rule.split()
    g_vector = []
    for gene in vector:
        g_vector.append( f'G_{gene}' )
    for gene in g_vector:
        if gene in fpkmDic:
            boolean_list.append('True')
        else:
            boolean_list.append('False')
    return boolean_list


##############################################################################
### Loading FPKM data
def loadFPKM(fpkm,condition,shuffle=False,shuffledFPKM=pd.DataFrame()):
    ##########################################################
    ## Gets gene IDs and their expression values
    genes=fpkm["Gene_ID"]
    if shuffle:
        fpkms=fpkm["Expression"].sample(frac=1).reset_index(drop=True) 
    else:
        fpkms=fpkm["Expression"]
    shuffledFPKM["Expression"] = fpkms
    ##########################################################
    ## Creates a dictionary gene_ID -> Expression value
    fpkmDic = {}
    for i in range(len(fpkms)): # Run over each line in fpkm file
        # 1. Get gene id and fpkm values for each line
        name = 'G_'+str(genes[i])
        fpkm = fpkms[i]
        if type(fpkm) == np.float64:
            fpkmDic[name] = fpkm
    ##########################################################
    ## Capping at 95%
    cap = np.percentile( list(fpkmDic.values()), 95)
    for i in fpkmDic:
        if fpkmDic[i]>cap:
            fpkmDic[i] = cap
    return(fpkmDic,shuffledFPKM)

##############################################################################
### Loading FPKM data for Homo sapiens. 
def reloadFPKMHsapiens(fpkmDic, model):
    newfpkmDic = {}
    for gene in model.genes:
        if not 'G_'+gene.name in fpkmDic: continue
        fpkm = fpkmDic['G_'+gene.name]
        gene = 'G_'+gene.id
        newfpkmDic[gene] = fpkm
    return(newfpkmDic)

##############################################################################
# UPDATE MODEL
def updateModel(model_default,mediumFile):
    model=model_default.copy()
    ##########################################################
    ## Add 'R_' to reactions names
    for reaction in model.reactions:
        reaction.id = 'R_'+reaction.id
    ##########################################################        
    ## Opening the model: exchange reactions
    for rxn in model.reactions: 
        if (rxn.lower_bound<0 and rxn.upper_bound>0):
            rxn.bounds = (-1000,1000)
        if (rxn.lower_bound>=0 and rxn.upper_bound>0):
            rxn.bounds = (0,1000)
        if (rxn.lower_bound<0 and rxn.upper_bound<=0):
            rxn.bounds = (-1000,0)
    ##########################################################
    ## Set culture medium
    #############################################
    if mediumFile != 'NA':
        #####################
        # load medium
        medium = pd.read_csv(mediumFile,sep="\t", lineterminator='\n')
        #####################
        # set bounds of exchange reactions to (0, 1000)
        for reaction in model.exchanges:
            reaction.bounds = (0, 1000)
        #####################
        # add culture medium
        for reaction in medium['Reaction_ID']:
            if 'R_'+reaction in model.reactions:
                model.reactions.get_by_id('R_'+reaction).lower_bound = -1000
        
    return(model)

##############################################################################
## Obtains a median value of the expression of genes associated with metabolism
def getEg(model,fpkmDic):
    g_metab = [] # gene expression of reactions partakin in the metabolism
    for i, reaction in enumerate(model.reactions):
        ##########################################################
        ## Gets the GPR and a boolean list with known or unknown genes
        rule = reaction.gene_reaction_rule
        boolean_list = booleanVectorRule(rule,fpkmDic)
        ##########################################################
        ## Gets the expression value 'g'
        if not ('False' in boolean_list or rule == ''): # get 'g' for reaction with GPR.
            g = getG(rule, fpkmDic)
            g_metab.append(g+1e-8)
    ##############################################################
    ## Obtains a median value
    E_g = np.median(g_metab)
    return(E_g,g_metab)

##############################################################################
### SAVE PRIMAL VALUES
def getPrimalValues(model):
    ##########################################################
    ### Model optimize: save fluxes and primal values of variables
    sol = model.optimize()
    fba_primal = {}
    for reaction in model.reactions:
        f_name = reaction.id
        r_name = reaction.reverse_id
        fba_primal[f_name] = eval ('model.variables.'+f_name+'.primal')
        fba_primal[r_name] = eval ('model.variables.'+r_name+'.primal')
    return(fba_primal)
##############################################################################
### Save forward and reverse variables
def getFowardReverse(model):
    v_vars, rev_vars = [], []
    for reaction in model.reactions:
        v_vars.append(reaction.id)
        rev_vars.append(reaction.reverse_id)  
    return(v_vars,rev_vars)


#################################################################################
### Variables and objective function: CasADI object
def setVariables(model,fpkmDic):
    v = vertcat() # saves the total of variables of the model. Used to """nlp['x']"""
    v_dic = {}
    v_fpkm = {} # 
    ubx, lbx = [],[]
    ##############################################################
    ## Gets the median value of 'g'
    E_g,g_metab = getEg(model,fpkmDic)     
    
    for i, reaction in enumerate(model.reactions):
        ##########################################################
        ## Gets the GPR and a boolean list with known or unknown genes
        rule = reaction.gene_reaction_rule # gene reaction rule
        boolean_list = booleanVectorRule(rule,fpkmDic) # useful to discriminate between genes with known fkpm.
        ##########################################################
        ## Gets the expression value 'g'
        # get 'g' for reaction with GPR.
        if not ('False' in boolean_list or rule == ''): 
            g = getG(rule, fpkmDic)+1e-8#*1e-8
        # set 'g' (median value) for reaction without GPR. 
        else:
            g = E_g
        ##########################################################
        ## Set forward and reverse variables as a CasADI object
        # forward
        var_name = reaction.id
        expression = var_name+' = SX.sym("'+var_name+'")'
        exec(expression, globals())
        vf = eval(var_name)
        v = vertcat(v, vf)
        v_dic[reaction.id]=vf
        ubx.append(reaction.upper_bound)    
        lbx.append(0.0)

        # reverse
        var_name_reverse = reaction.reverse_id
        expression = var_name_reverse+' = SX.sym("'+var_name_reverse+'")'
        exec(expression, globals())
        vr = eval(var_name_reverse)
        v = vertcat(v,vr)
        v_dic[reaction.reverse_id]=vr
        ubx.append(-reaction.lower_bound)
        lbx.append(0.0)

        v_fpkm[var_name] = g
        v_fpkm[var_name_reverse] = g
        ##########################################################
        ## Define a objective function
        for name in [vf,vr]:
            if i == 0:
                v_ViLogVi = ( (name)+1e-8 )*log( (name)+1e-8 ) # 1.1
                v_VilogQi = ( (name)+1e-8 )*log( g ) # 2.1
            else:
                v_ViLogVi += ( (name)+1e-8 )*log( (name)+1e-8 ) # 1.1
                v_VilogQi += ( (name)+1e-8 )*log( g ) # 2.1            
    ##############################################################
    ## Set objetive function
    f = (v_ViLogVi) - (v_VilogQi)
    return(v,v_dic,lbx,ubx,f)

#################################################################################
### Define a sumV
def getSumV(v):
    for i in range(v.shape[0]): # VARIABLES
        name = v[i]
        if i == 0:
            sumVi = name
        else:
            sumVi += name     
    return(sumVi)

#################################################################################
### Creating constraints
def createConstraints(model,k,v_dic,sumVi):
    g = vertcat()
    lbg,ubg=[],[]
    ##############################################################
    ## Gets the name of the forward/reverse variables
    v_vars, rev_vars = getFowardReverse(model)
    ##############################################################
    ## Defines constraints
    for met in model.metabolites:
        ##########################################################
        ## Gets constraint for a one metabolite
        constraint = str(met.constraint.expression).replace('+ ','+').replace('- ','-')
        ##########################################################
        ## Reconstruct the constraint as a CasADI object
        for i, field in enumerate(constraint.split()):
            if i == 0:
                tmp_constraint = eval(field)
            else:
                tmp_constraint += eval(field)
                
        ##########################################################
        ## Adds constraint to CasADI
        g = vertcat(g,tmp_constraint)
        lbg.append(0)
        ubg.append(0)
    ##############################################################
    ## SumV constraint
    g = vertcat(g,sumVi)
    lbg.append( k )
    ubg.append( k )
    
    return(ubg,lbg,g)

#################################################################################
### OPTIMIZATION
def optPheFlux(model,fpkmDic,k,init_time):
    ##############################################################
    ## Sets variables, sumV and constraints
    v,v_dic,lbx,ubx,f = setVariables(model,fpkmDic)
    sumVi = getSumV(v)
    ubg,lbg,g = createConstraints(model,k,v_dic,sumVi)    
    print('')
    ##############################################################
    # Non-linear programming
    nlp = {}     # NLP declaration
    nlp['x']=  v # variables
    nlp['f'] = f # objective function
    nlp['g'] = g # constraints
    ##############################################################
    # Create solver instance
    options={"ipopt":{"print_level":3}}
    F = nlpsol('F','ipopt',nlp,options)
    ##############################################################
    # Solve the problem using a guess
    fba_primal = getPrimalValues(model)
    x0=[]
    for i in range(v.shape[0]): # VARIABLES
        x0.append(fba_primal[str(v[i])])       
    ##############################################################
    ## Solver
    start = time.time()
    sol=F(x0=x0,ubg=ubg,lbg=lbg,lbx=lbx,ubx=ubx)
    final = time.time()
    total_time = final - init_time
    optimization_time = final - start
    status = F.stats()['return_status']
    success = F.stats()['success']
    ##############################################################
    ## Save data as Pandas Series
    PheFlux = sol['x']
    PheFlux_fluxes = {}
    for num, i in enumerate (range(0, v.shape[0] , 2)):
        name = str(v[i])
        reaction_flux = ( PheFlux[i] - PheFlux[i+1] ) # (forward - reverse)
        PheFlux_fluxes[name] =  float(reaction_flux)
    PheFlux_fluxes = pd.Series(PheFlux_fluxes)

    return(PheFlux_fluxes,optimization_time,total_time,status,success, lbx, ubx)

#################################################################################
### Table of times and variable numbers for networks
def recordTable(record ,condition, lbx, ubx, time, status):   
    variables = 0
    for i in range(len(lbx)):
        if lbx[i] != ubx[i]:
            variables += 1
    
    if record.shape == (0,0):
        record = pd.DataFrame(columns=['Condition', 'N° variables', 'Time', 'Status'])
        
    record.loc[record.shape[0]] = [condition, variables, time, status]
    return (record)

#################################################################################
### print the time
def actual_time():
    now = datetime.now(timezone.utc)
    return now.isoformat()




import polars as pl
from collections import namedtuple
from rich import print
from cobra.core.model import Model as SBMLModel
from collections import namedtuple
from casadi import nlpsol
import time
import pandas as pd
import polars as pl
from pheflux.set_variables import set_variables, CasADIData
from pheflux.sum_v import getSumV
from pheflux.create_constraints import createConstraints

Fluxes = namedtuple("Fluxes",[
    "series",
    "optimization_time",
    "total_time",
    "status",
    "success_flag",
    "bounds"
])


def fill_unknown(df:pl.DataFrame)->pl.DataFrame:
    exists = df.filter(pl.col('Gene_ID') == "G_UNKNOWN").count().to_dict()["Gene_ID"][0] > 0

    if not exists:
        df = df[["Gene_ID", "Expression"]]
        average_expression = df.select(
            pl.col("Expression").mean()
        ).item()
        item = {"Gene_ID":"G_UNKNOWN",
                "Expression":average_expression}
        new_row = pl.DataFrame([item])
        df = df.vstack(new_row)
        return df
    else:
        return fpkm

def opt_phe_flux(
        model:SBMLModel,
        fpkm:pl.DataFrame,
        init_time:float,
        k:float=1000,
)->Fluxes:

    """
    Input:
        model (cobra.Model): Metabolic model.
        fpkmDic (dict): Dictionary with gene expression values.
        k (float): SumV constraint value.
        init_time (float): Initial time for measuring optimization time.
    
    Output:
        tuple: Series with fluxes, optimization time, total time, status, success flag, and lists of lower and upper bounds.
    """
    fpkm = fill_unknown(fpkm)
    casadi_data= set_variables(model,fpkm)
    sumVi = getSumV(casadi_data.casadi)
    ubg, lbg, g = createConstraints(
        model,
        casadi_data.names,
        k,
        casadi_data.variables,
        sumVi, 
        casadi_data.sx_variables)    

    #v,v_dic,lbx,ubx,f 
    v = casadi_data.casadi
    f = casadi_data.objetive_fn
    lbx = casadi_data.lower_bounds
    ubx = casadi_data.upper_bounds
    ##############################################################
    # Non-linear programming
    nlp = {}     # NLP declaration
    nlp['x']=  v # variables
    nlp['f'] = f # objective function
    nlp['g'] = g # constraints

    ##############################################################
    # Create solver instance, define Fcas optimization function 
    options = {"ipopt":{"print_level":3}}
    Fcas = nlpsol('F','ipopt',nlp,options)
    ##############################################################
    ##############################################################
    # Solve the problem using a guess
    fba_primal = casadi_data.primal
    x0=[]
    for i in range(v.shape[0]): # VARIABLES
        x0.append(fba_primal[str(v[i])])       
    ## Solver
    start = time.time()
    sol=Fcas(x0=x0,ubg=ubg,lbg=lbg,lbx=lbx,ubx=ubx)
    final = time.time()
    total_time = final - init_time
    optimization_time = final - start
    status = Fcas.stats()['return_status']
    success = Fcas.stats()['success']
    ##############################################################
    ## Save data as Pandas Series
    PheFlux = sol['x']
    PheFlux_fluxes = {}
    dataset = []
    for num, i in enumerate (range(0, v.shape[0] , 2)):
        name = str(v[i])
        reaction_flux = ( PheFlux[i] - PheFlux[i+1] ) # (forward - reverse)
        item = {"name":name, "value": float(reaction_flux)}
        dataset.append(item)

    PheFlux_fluxes = pl.DataFrame(dataset)

    return PheFlux_fluxes,optimization_time,total_time,status,success, lbx, ubx



import polars as pl
import ply.lex as lex

def boolean_vector_rule(
        tokens:list[lex.LexToken],
        fpkm:pl.DataFrame
)->list[bool]:
    """
    Input:
        rule (str): Gene-protein rule as a string.
        fpkmDic (dict): Dictionary with gene expression values.
    
    Output:
        list: List of boolean values indicating if the FPKM values for the genes are known.
    """
    try:
        # set(tokens) saves repetition 
        genes_bools = {}
        gen_tokens = {t for t in tokens if t.type=="GENE"}
        for token in gen_tokens:
            key = f"G_{token.value.upper()}"
            genes_bools[key] = fpkm["Gene_ID"].is_in([key]).any() 
        return genes_bools
    except Exception as e:
        raise e



from pathlib import Path

from pydantic import BaseModel
from pydantic import validator
from pydantic.dataclasses import dataclass

from typing import ClassVar, List, TypeVar, Optional,Any
from cobra.core.model import Model as SBMLModel
import polars as pl
import cobra
from concurrent.futures import ThreadPoolExecutor, as_completed

# Organism->Condition
@dataclass(config=dict(arbitrary_types_allowed=True))
class OrganismItem:
    organism: str
    condition:str
    gene_exp_file:Path
    medium:Path
    network:Path

    gene_exp_data: Optional[pl.DataFrame] = None
    medium_data: Optional[pl.DataFrame] = None
    network_data: Optional[SBMLModel] = None  # Assuming SBML data as cobra.Model

    allowed_gene_exp_extensions: ClassVar[List[str]] = ['.csv', '.txt']
    allowed_medium_extensions: ClassVar[List[str]] = ['.csv']
    allowed_network_extensions: ClassVar[List[str]] = ['.xml']


    def activate(self):
        # Start reading files in parallel
        self.future_to_file = {
            'gene_exp_file': None,
            'medium': None,
            'network': None
        }
        with ThreadPoolExecutor() as executor:
            self.future_to_file['gene_exp_file'] = executor.submit(self.read_gene_file)
            self.future_to_file['medium'] = executor.submit(self.read_medium_file)
            self.future_to_file['network'] = executor.submit(self.read_network_file)
        
        self.wait_for_files()


    def wait_for_files(self):
        """
        Espera hasta que todos los archivos hayan sido leídos.
        """
        for file_type, future in self.future_to_file.items():
            try:
                data = future.result()
                if file_type == 'gene_exp_file':
                    self.gene_exp_data = data
                elif file_type == 'medium':
                    self.medium_data = data
                elif file_type == 'network':
                    self.network_data = data
            except Exception as exc:
                print(f"{file_type} generated an exception: {exc}")

    @validator('gene_exp_file')
    def check_gene_exp_file(cls, value: Path) -> Path:
        if value.suffix not in cls.allowed_gene_exp_extensions:
            raise ValueError(f'gene_exp_file must have one of the following extensions: {cls.allowed_gene_exp_extensions}')
        if not value.exists():
            raise ValueError(f'gene_exp_file path does not exist: {value}')
        return value
    @validator('medium')
    def check_medium(cls, value: Path) -> Path:
        if value.suffix not in cls.allowed_medium_extensions:
            raise ValueError(f'medium must have one of the following extensions: {cls.allowed_medium_extensions}')
        if not value.exists():
            raise ValueError(f'medium path does not exist: {value}')
        return value

    @validator('network')
    def check_network(cls, value: Path) -> Path:
        if value.suffix not in cls.allowed_network_extensions:
            raise ValueError(f'network must have one of the following extensions: {cls.allowed_network_extensions}')
        if not value.exists():
            raise ValueError(f'network path does not exist: {value}')
        return value


    def read_gene_file(self) -> pl.DataFrame|None:
        """
        Lee el archivo gene_exp_file basado en su extensión y retorna un DataFrame de Polars.
        """
        try:
            if self.gene_exp_file.suffix in self.allowed_gene_exp_extensions:
                df = pl.read_csv(self.gene_exp_file, separator='\t')
                return df
        except Exception as e:
            print(f"Error al leer gene_exp_file: {e}")
        return None

    def read_medium_file(self) -> pl.DataFrame|None:
        """
        Lee el archivo medium basado en su extensión y retorna un DataFrame de Polars.
        """
        try:
            if self.medium.suffix in self.allowed_medium_extensions:
                df = pl.read_csv(self.medium, separator='\t')
                return df
        except Exception as e:
            print(f"Error al leer medium: {e}")
        return None

    def read_network_file(self) -> SBMLModel|None:
        """
        Lee el archivo network basado en su extensión y retorna un objeto cobra.Model.
        """
        try:
            if self.network.suffix in self.allowed_network_extensions:
                model = cobra.io.read_sbml_model(self.network)
                return model
        except Exception as e:
            print(f"Error al leer network: {e}")
        return None



from casadi import SX

def getSumV(casadi:SX)->SX:
    """
    Input:
        v (CasADI object): Vector of variables.
    
    Output:
        CasADI object: Sum of all variables.
    """
    n = casadi.shape[0]
    for i in range(n): # VARIABLES
        name = casadi[i]
        if i == 0:
            sumVi = name
        else:
            sumVi += name     
    return sumVi 



import time
import json
from pydantic import BaseModel
from typing import List
import numpy as np

class MeasureTime(BaseModel):
    start_time: float
    end_time: float
    opt_time: List[float]
    t_time: List[float]

    def generate_report(self) -> str:
        process_time = self.end_time - self.start_time
        if not self.opt_time:
            self.opt_time=[-1]
        if not self.t_time:
            self.t_time=[-1]
        report = {
            'Average time per optimization': np.mean(self.opt_time),
            'Average time per condition': np.mean(self.t_time),
            'Total process time (minutes)': process_time / 60,
            'Total process time (hours)': process_time / 3600
        }
        return json.dumps(report, indent=4)


